/**
 ** search.c - helper program to search a binary file for a set
 **            sequence of hex values. It will display the
 **            offset along with the matched pattern.
 **
 **     usage: the program takes a pattern of space-separated
 **            byte values (0-9a-f), along with a half-byte
 **            "any value" regex: '.'. Source data is given
 **            via STDIN (typically via the receiving end of
 **            a pipe.
 **
 **     notes: logic isn't thorough; command-line processing
 **            assumes user always does the right thing.
 **
 **            program logic assumes only 1 byte values.
 **
 **   compile: gcc -o search search.c -Wall
 **   execute: cat file | search 'de ad be ef'
 **        or: echo "de ad be ef 00 01 5a" | search 'de ad'
 **/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define  EVEN    0
#define  ODD    1

typedef unsigned char      uc;
typedef   signed short int ssi;
typedef unsigned short int usi;

int main(int argc, char **argv)
{
    //////////////////////////////////////////////////////////////////
    //
    // A few variables to help along the way
    //
    FILE *input                            = NULL;
    int   index                            = 0;
    int   pos                              = 0;
    uc    value                            = 0;
    uc    upper_nibble                     = 0;
    uc    lower_nibble                     = 0;
    uc    length                           = 0;
    uc    regex                            = 0;
    usi  *data                             = NULL;
    uc   *match                            = NULL;

    //////////////////////////////////////////////////////////////////
    //
    // Not enough arguments (or too many) were given
    //
    if ((argc                             <  2) ||
        (argc                             >  3))
    {
        fprintf(stderr, "Error: invalid argument count!\n");
        fprintf(stderr, "Usage: %s PATTERN [FILE]\n", *(argv+0));
        exit(1);
    }

    //////////////////////////////////////////////////////////////////
    //
    // Quick parse of pattern to get rough initial length
    //
    while (*(*(argv+1)+index)             != '\0')
    {
        if (!*(*(argv+1)+index)           == ' ')
        {
            length                         = length + 1;
        }
        index                              = index  + 1;
    }

    if ((length % 2)                      == 1)
    {
        fprintf(stderr, "Error: Malformed PATTERN '%s'\n", *(argv+1));
        exit(1);
    }

    //////////////////////////////////////////////////////////////////
    //
    // Check for optional file argument (read from STDIN otherwise)
    //
    if (argc                              == 2)
    {
        input                              = stdin;
    }
    else
    {
        input                              = fopen(*(argv+2), "r");
        if (input                         == NULL)
        {
            fprintf(stderr, "Error: Could not open '%s' for reading!\n",
                            *(argv+2));
            exit(1);
        }
    }

    /////////////////////////////////////////////////////////////////
    //
    // Since we operate in groups of 2 (for a whole byte), cut our
    // length value in half
    //
    length                                 = length / 2;

    /////////////////////////////////////////////////////////////////
    //
    // Allocate storage for our data and matched patterns
    //
    data                                   = (usi *)malloc(sizeof(usi)*length);
    match                                  = (uc *)malloc(sizeof(uc)*length);

    /////////////////////////////////////////////////////////////////
    //
    // Parse pattern for individual values, and construct values
    //
    index                                  = 0;
    for (pos = 0; pos < strlen(*(argv+1)); pos += 2)
    {
        regex                              = 0;
        upper_nibble                       = *(*(argv+1)+pos);

		/////////////////////////////////////////////////////////////
		//
		// Check to see if it is a number, adjust accordingly
		//
        if ((upper_nibble                 >= '0') &&
            (upper_nibble                 <= '9'))
        {
            upper_nibble                   = upper_nibble - 48;
        }

		/////////////////////////////////////////////////////////////
		//
		// Check to see if it is a lowercase letter
		//
        else if ((upper_nibble            >= 'a') &&
                 (upper_nibble            <= 'f'))
        {
            upper_nibble                   = upper_nibble - 87;
        }

		/////////////////////////////////////////////////////////////
		//
		// Check to see if it is an uppercase letter
		//
        else if ((upper_nibble            >= 'A') &&
                 (upper_nibble            <= 'F'))
        {
            upper_nibble                   = upper_nibble - 55;
        }

		/////////////////////////////////////////////////////////////
		//
		// Check to see if it is a regular expression
		//
        else if (upper_nibble             == '.')
        {
			regex                          = 1;
            upper_nibble                   = 0;
        }

		/////////////////////////////////////////////////////////////
		//
		// If none of these, we have an error
		//
		else
		{
			fprintf(stderr, "Error: Invalid pattern symbol '%c' (%hhu)\n",
							upper_nibble,
							upper_nibble);
			exit(1);
		}

		/////////////////////////////////////////////////////////////
		//
		// Position upper_nibble to the upper-order bits of byte,
		// and scrub any errant values not in those bit positions
		//
		upper_nibble                       = upper_nibble << 4;
		upper_nibble                       = upper_nibble &  0x00F0;

        lower_nibble                       = *(*(argv+1)+(pos+1));

		/////////////////////////////////////////////////////////////
		//
		// Check to see if it is a number, adjust accordingly
		//
        if ((lower_nibble                 >= '0') &&
            (lower_nibble                 <= '9'))
        {
            lower_nibble                   = lower_nibble - 48;
        }

		/////////////////////////////////////////////////////////////
		//
		// Check to see if it is a lowercase letter
		//
        else if ((lower_nibble            >= 'a') &&
                 (lower_nibble            <= 'f'))
        {
            lower_nibble                   = lower_nibble - 87;
        }

		/////////////////////////////////////////////////////////////
		//
		// Check to see if it is an uppercase letter
		//
        else if ((lower_nibble            >= 'A') &&
                 (lower_nibble            <= 'F'))
        {
            lower_nibble                   = lower_nibble - 55;
        }

		/////////////////////////////////////////////////////////////
		//
		// Check to see if it is a regular expression
		//
        else if (lower_nibble             == '.')
        {
			regex                          = regex | 2;
            lower_nibble                   = 0;
        }

		/////////////////////////////////////////////////////////////
		//
		// If none of these, we have an error
		//
		else
		{
			fprintf(stderr, "Error: Invalid pattern symbol '%c' (%hhu)\n",
							lower_nibble,
							lower_nibble);
			exit(1);
		}

		/////////////////////////////////////////////////////////////
		//
		// Scrub any errant values not in those bit positions
		//
		lower_nibble                       = lower_nibble &  0x000F;

		/////////////////////////////////////////////////////////////
		//
		// Pack the values into one element in data array, based on
		// whether or not there is a regex, and where it is
		//
		*(data+index)                      = 0;
		if (regex                         == 0)
		{
			*(data+index)                  = upper_nibble | lower_nibble & 0x00FF;
		}
		else if (regex                    == 1)
		{
			*(data+index)                  = upper_nibble | lower_nibble | 0xF000;
		}
		else if (regex                    == 2)
		{
			*(data+index)                  = upper_nibble | lower_nibble | 0x0F00;
		}
		else
		{
			*(data+index)                  = upper_nibble | lower_nibble | 0xFF00;
		}

		index                              = index + 1;

		/////////////////////////////////////////////////////////////
		//
		// Support space-less input (experimental)
		//
		if (*(*(argv+1)+(pos+2))          == 0x20)
			pos                            = pos + 1;
	}





























    //////////////////////////////////////////////////////////////////
    //
    // Determine operating mode (command-line args (this) XOR STDIN)
    //
    if (argc                              == 3)
    {
        //////////////////////////////////////////////////////////////
        //
        // Iterate through each bit
        //
        while (*(*(argv+2)+index)         != '\0')
        {
            //////////////////////////////////////////////////////////
            //
            // Check for a '1' or a '0'
            //
            if (*(*(argv+2)+index)        == '1')
                value                      = value + 1;

            index                          = index + 1;
        }
    }

    //////////////////////////////////////////////////////////////////
    //
    // STDIN mode
    //
    else if (argc                         == 2)
    {
        //////////////////////////////////////////////////////////////
        //
        // Iterate through each bit of the input
        //
        do
        {
            //////////////////////////////////////////////////////////
            //
            // Grab a 1 byte hex value from STDIN
            //
            index                          = fscanf(stdin, "%c", &bit);

            //////////////////////////////////////////////////////////
            //
            // If EOF character, break out
            //
            if ((index                    == -1) ||
                (bit                      == '\n'))
            {
                break;
            }

            //////////////////////////////////////////////////////////
            //
            // Check for invalid input and bail out if so
            //
            if ((bit                      != '0') &&
                (bit                      != '1'))
            {
                fprintf(stderr, "*** Error: Invalid input value '%c'!\n", bit);
                exit(0);
            }

            //////////////////////////////////////////////////////////
            //
            // Iterate through each bit
            //
            if (bit                       == '1')
            {
                //////////////////////////////////////////////////////
                //
                // Check for a '1' or a '0'
                //
                if (bit                   == '1')
                    value                  = value + 1;
            }

        } while (!feof(stdin));
    }

    //////////////////////////////////////////////////////////////////
    //
    // Verify and display the parity
    //
    if ((value % 2)                       == parity)
        fprintf(stderr, "true");
    else
        fprintf(stderr, "false");

    //////////////////////////////////////////////////////////////////
    //
    // Newline at the end
    //
    fprintf(stdout, "\n");

    return(0);
}
